@@ -1,3 +1,5 @@
+#include "devfuns.h"
+#include "raster.h"
 #include "outline.h"
 #include "util.h"
 
@@ -5,42 +7,49 @@ void showtextMetricInfo(int c, const pGEcontext gc, double* ascent, double* desc
 {
     FT_Face face = GetFTFace(gc);
     FT_Error err;
-    double fontSize = gc->ps * gc->cex;
-    double ratio = fontSize / face->units_per_EM;
+    
+    /* Font size in points (1/72 inches) */
+    double font_size = gc->ps * gc->cex;
+    /* Metrics given by FreeType are represented in EM units when
+       we set FT_LOAD_NO_SCALE.
+       Therefore, we first transform EM units into points. */
+    double pts_per_EM_unit = font_size / face->units_per_EM;
+    /* Then we further convert points to device units.
+       Device unit can be in points (usually vector graphics),
+       or in pixels (usually bitmap graphics). */
+    double dev_units_per_EM_unit = pts_per_EM_unit * GetDevUnitsPerPoint();
   
-    if(c == 0) c = 77;
-    if(c < 0)
-    {
-        c = -c;
-    }
+    if(c == 0) c = 77;  /* Letter 'M' */
+    if(c < 0)  c = -c;
     
     /* c is the unicode of the character */
-    FT_Set_Char_Size(face, 0, fontSize * 64, 72, 0);
     err = FT_Load_Char(face, c, FT_LOAD_NO_SCALE);
     if(err)
     {
-        errorcode(err);
+        FTError(err);
         *ascent = *descent = *width = 0.0;
         return;
     }
     
-    *ascent = face->glyph->metrics.horiBearingY * ratio;
-    *descent = face->glyph->metrics.height * ratio - *ascent;
-    *width = face->glyph->metrics.horiAdvance * ratio;
+    *ascent = face->glyph->metrics.horiBearingY * dev_units_per_EM_unit;
+    *descent = face->glyph->metrics.height * dev_units_per_EM_unit - *ascent;
+    *width = face->glyph->metrics.horiAdvance * dev_units_per_EM_unit;
 }
 
 double showtextStrWidthUTF8(const char *str, const pGEcontext gc, pDevDesc dd)
 {
     /* Convert UTF-8 string to Unicode array */
-    int maxLen = strlen(str);
+    int max_len = strlen(str);
     unsigned int *unicode =
-        (unsigned int *) calloc(maxLen + 1, sizeof(unsigned int));
-    int len = utf8toucs4(unicode, str, maxLen);
+        (unsigned int *) calloc(max_len + 1, sizeof(unsigned int));
+    int len = utf8toucs4(unicode, str, max_len);
     
     FT_Face face = GetFTFace(gc);
     FT_Error err;
-    double fontSize = gc->ps * gc->cex;
-    double ratio = fontSize / face->units_per_EM;
+    
+    double font_size = gc->ps * gc->cex;
+    double pts_per_EM_unit = font_size / face->units_per_EM;
+    double dev_units_per_EM_unit = pts_per_EM_unit * GetDevUnitsPerPoint();
     
     double width = 0.0;
     int i;
@@ -50,16 +59,42 @@ double showtextStrWidthUTF8(const char *str, const pGEcontext gc, pDevDesc dd)
         err = FT_Load_Char(face, unicode[i], FT_LOAD_NO_SCALE);
         if(err)
         {
-            errorcode(err);
+            FTError(err);
             continue;
         }
-        width += face->glyph->metrics.horiAdvance * ratio;
+        width += face->glyph->metrics.horiAdvance * dev_units_per_EM_unit;
     }
 
     return width;
 }
 
-void showtextTextUTF8(double x, double y, const char *str, double rot, double hadj, const pGEcontext gc, pDevDesc dd)
+void showtextTextUTF8Raster(double x, double y, const char *str, double rot, double hadj, const pGEcontext gc, pDevDesc dd)
+{
+    /* Convert UTF-8 string to Unicode array */
+    int max_len = strlen(str);
+    unsigned int *unicode =
+        (unsigned int *) calloc(max_len + 1, sizeof(unsigned int));
+    int len = utf8toucs4(unicode, str, max_len);
+    
+    /* raster() rotates around the bottom-left corner,
+       and text() rotates around the center indicated by hadj. */
+    int trans_sign = dd->bottom > dd->top ? -1: 1;
+    double trans_X, trans_Y;
+    
+    /* Calculate pixel size */
+    int px = (int) (gc->ps * gc->cex * GetDevUnitsPerPoint() + 0.5);
+    
+    /* Get raster data */
+    RasterData *rd = GetStringRasterImage(unicode, len, px, px,
+        rot * DEG2RAD, hadj, gc, &trans_X, &trans_Y);
+
+    dd->raster(rd->data, rd->ncol, rd->nrow,
+               x - trans_X, y - trans_sign * trans_Y,
+               rd->ncol, -rd->nrow, 0.0, FALSE, gc, dd);
+    FreeRasterData(rd);
+}
+
+void showtextTextUTF8Polygon(double x, double y, const char *str, double rot, double hadj, const pGEcontext gc, pDevDesc dd)
 {
     /* Convert UTF-8 string to Unicode array */
     int maxLen = strlen(str);
@@ -103,14 +138,14 @@ void showtextTextUTF8(double x, double y, const char *str, double rot, double ha
         err = FT_Load_Char(face, unicode[i], FT_LOAD_NO_SCALE);
         if(err)
         {
-            errorcode(err);
+            FTError(err);
             continue;
         }
         outline = face->glyph->outline;
         err = FT_Outline_Decompose(&outline, funs, &data);
         if(err)
         {
-            errorcode(err);
+            FTError(err);
             ArrayDestroy(data.outline_x);
             ArrayDestroy(data.outline_y);
             data.outline_x = ArrayNew(100);
